/* ctime.c - a Ruby interface to Corporate Time
 * 
 * $Id: ctime.c,v 1.88 2006/05/25 19:23:14 ianmacd Exp $
 *
 * Author: Ian Macdonald <ian@caliban.org>
 *
 * Copyright (C) 2002-2006 Ian Macdonald
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software Foundation,
 *   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <ruby.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Required for ctapi.h */
#ifdef WIN32
#define CAPI_FUNC(a)  __declspec(dllimport) a
#else
#define CAPI_FUNC(a) a
#endif

#include <ctapi.h>
#include "capilayer.h"

#define CTIME_VERSION   "0.4.9"
#define CTIME_STRFTIME	"%Y%m%dT%H%M00Z"
#define MAX_HANDLES	32

/* Template for an MIME wrapper iCalendar event. */
static const char *MIME_HEADER = {
  "MIME-Version: 1.0\r\n"
    "Content-Type: text/calendar\r\n"
    "Content-Transfer-Encoding: quoted-printable\r\n\r\n" "%s"
};

static const char *ICAL_TEMPLATE = {
  "BEGIN:VCALENDAR\r\n"
    "VERSION:2.0\r\n"
    "PRODID://Steltor//CAPI//EN\r\n"
    "BEGIN:VEVENT\r\n" "PRIORITY:5\r\n"
    "%s" "END:VEVENT\r\n" "END:VCALENDAR\r\n"
};

int agenda_count = 0;

/* Leaving an extra slot for null-terminating the array. */
CAPIHandle agenda[MAX_HANDLES + 1] = { NULL };

/* The name of the main class. */
VALUE cCTime;

/* Other classes to which we refer. */
VALUE cEvent;
VALUE cAgenda;

/* These are exceptions in the CTime class. */
VALUE eCTime_CAPIError;
VALUE eCTime_AgendaError;
VALUE eCTime_EventError;
VALUE eCTime_DesignationError;


/* Get a string description of the given CAPIStatus and print it to debug */
static const char *
ct_status_update (CAPIStatus c_status)
{
  const char *error_name = NULL;

  capi_get_status_string (c_status, &error_name);
  rb_ctime_debug ("Status: %s\n", error_name);

  return error_name;
}


/* Check that a string property of an object is not nil and not empty */
static const char *
ct_get_strprop (VALUE object, const char *propname)
{
  const char *strprop = NULL;
  VALUE prop = rb_iv_get (object, propname);

  if (prop == Qnil)
    return NULL;

  strprop = STR2CSTR (prop);

  if (NULL != strprop && strprop[0] != 0)
    return strprop;

  return NULL;
}


/* Allocate and wrap a CAPI session. */
static VALUE
ct_allocate (VALUE klass)
{
  CAPISession *c_session = NULL;

  rb_ctime_debug ("Allocating a CAPI session...\n");

  /* No need to define a function to free this, as CTime#disconnect will
     take care of it for us. */
  return Data_Make_Struct (klass, CAPISession, NULL, NULL, c_session);
}


/* Create a memory stream. */
static CAPIStream *
ct_make_mem_stream (CAPISession in_session,
		    CAPIStream * out_stream,
		    const char *in_buffer, const char **out_buffer)
{
  CAPIStatus c_status = CAPI_STAT_OK;
  const char *error = NULL;

  rb_ctime_debug ("Creating memory stream...\n");
  c_status =
    capi_create_memory_stream (in_session, out_stream, in_buffer, out_buffer);
  error = ct_status_update (c_status);

  if (c_status != CAPI_STAT_OK)
    rb_raise (eCTime_CAPIError, "Failed to create memory stream: %s", error);

  return out_stream;
}


/* Destroy a memory stream. */
static void
ct_kill_mem_stream (CAPISession in_session, CAPIStream * out_stream)
{
  CAPIStatus c_status = CAPI_STAT_OK;
  const char *error = NULL;

  rb_ctime_debug ("Destroying memory stream...\n");
  c_status = capi_destroy_stream (in_session, out_stream);
  error = ct_status_update (c_status);

  if (c_status != CAPI_STAT_OK)
    rb_raise (eCTime_CAPIError, "Failed to destroy memory stream: %s", error);

  return;
}


/* Report number of open agendas. */
static void
ct_report_agendas ()
{
  int i;

  if (agenda_count == 1)
    {
      rb_ctime_debug ("There is 1 open agenda:\n");
    }
  else if (agenda_count == 0)
    {
      rb_ctime_debug ("There are 0 open agendas.\n");
    }

  else
    {
      rb_ctime_debug ("There are %d open agendas:\n", agenda_count);
    }

  for (i = 0; agenda[i]; ++i)
    rb_ctime_debug ("agenda[%d] = %p\n", i, agenda[i]);
}


/* CTime#disconnect */
static VALUE
ct_quit (VALUE self)
{
  CAPISession *c_session = NULL;
  CAPIStatus c_status = CAPI_STAT_OK;

  Data_Get_Struct (self, CAPISession, c_session);

  if (c_session && agenda_count > 0)
    {
      ct_report_agendas ();
      rb_ctime_debug ("Destroying agenda handles...\n");
      c_status =
	capi_destroy_multiple_handles (*c_session, agenda, agenda_count);
      ct_status_update (c_status);
      agenda_count = 0;
      agenda[agenda_count] = NULL;
    }

  if (c_session)
    {
      rb_ctime_debug ("Disconnecting from server...\n");
      c_status = capi_disconnect (c_session);

      ct_status_update (c_status);
      if (c_status == CAPI_STAT_API_SESSION_NULL)
	rb_raise (eCTime_CAPIError, "Already disconnected from server");
    }

  return Qnil;
}


/* CTime#connect */
static VALUE
ct_connect (VALUE self, VALUE server, VALUE user, VALUE passwd)
{
  CAPISession *c_session = NULL;
  CAPIStatus c_status = CAPI_STAT_OK;

  rb_ctime_debug ("Connecting to %s...\n", STR2CSTR (server));
  Data_Get_Struct (self, CAPISession, c_session);
  c_status = capi_connect (STR2CSTR (server), STR2CSTR (user),
			   STR2CSTR (passwd), c_session);
  ct_status_update (c_status);

  if (c_status == CAPI_STAT_OK)
    {
      /* Yield connection to the block, if given. */
      if (rb_block_given_p ())
	rb_ensure (rb_yield, self, ct_quit, self);
      return self;
    }
  else if (c_status == CAPI_STAT_API_SESSION_NOTNULL)
    rb_raise (eCTime_CAPIError, "Connection to %s already established",
	      STR2CSTR (server));
  else if (c_status == CAPI_STAT_DATA_USERID_ID)
    rb_raise (eCTime_CAPIError,
	      "Can't log on as user %s. Does the user exist?",
	      STR2CSTR (user));
  else if (c_status == CAPI_STAT_SECUR_LOGON_AUTH)
    rb_raise (eCTime_CAPIError, "Incorrect password for %s", STR2CSTR (user));
  else if (c_status == CAPI_STAT_DATA_HOSTNAME_SERVER)
    rb_raise (eCTime_CAPIError, "Server %s not found", STR2CSTR (server));
  else
    rb_raise (eCTime_CAPIError, "Unknown failure when connecting to %s",
	      STR2CSTR (server));

  return Qnil;
}


/* CTime#capabilities */
static VALUE
ct_capabilities (VALUE self)
{
  const char *txt = NULL;
  const char *comma = ",";
  const char *comma_sp = ", ";
  CAPISession *c_session = NULL;
  VALUE capabilities;
  VALUE cap_list;

  rb_ctime_debug ("Fetching server capabilities...\n");
  Data_Get_Struct (self, CAPISession, c_session);
  capabilities = rb_hash_new ();

  capi_get_capabilities (*c_session, CAPI_CAPAB_AUTH, &txt);
  cap_list = rb_str_split (rb_str_new2 (txt), comma);
  rb_hash_aset (capabilities, rb_str_new2 ("authentication"), cap_list);

  capi_get_capabilities (*c_session, CAPI_CAPAB_COMP, &txt);
  cap_list = rb_str_split (rb_str_new2 (txt), comma);
  rb_hash_aset (capabilities, rb_str_new2 ("compression"), cap_list);

  capi_get_capabilities (*c_session, CAPI_CAPAB_ENCR, &txt);
  cap_list = rb_str_split (rb_str_new2 (txt), comma);
  rb_hash_aset (capabilities, rb_str_new2 ("encryption"), cap_list);

  capi_get_capabilities (*c_session, CAPI_CAPAB_MAXDATE, &txt);
  rb_hash_aset (capabilities, rb_str_new2 ("max_date"), rb_str_new2 (txt));

  capi_get_capabilities (*c_session, CAPI_CAPAB_CAPI_VERSION, &txt);
  rb_hash_aset (capabilities, rb_str_new2 ("capi_version"),
		rb_str_new2 (txt));

  capi_get_capabilities (*c_session, CAPI_CAPAB_VERSION, &txt);
  rb_hash_aset (capabilities, rb_str_new2 ("version"), rb_str_new2 (txt));

  capi_get_capabilities (*c_session, CAPI_CAPAB_SERVER_VERSION, &txt);
  rb_hash_aset (capabilities, rb_str_new2 ("server_version"),
		rb_str_new2 (txt));

  capi_get_capabilities (*c_session, CAPI_CAPAB_UNSUPPORTED_ICAL_PROP, &txt);
  cap_list = rb_str_split (rb_str_new2 (txt), comma_sp);
  rb_hash_aset (capabilities, rb_str_new2 ("unsupported_ical_prop"),
		cap_list);

  capi_get_capabilities (*c_session, CAPI_CAPAB_UNSUPPORTED_ICAL_COMP, &txt);
  cap_list = rb_str_split (rb_str_new2 (txt), comma);
  rb_hash_aset (capabilities, rb_str_new2 ("unsupported_ical_comp"),
		cap_list);

  return capabilities;
}


/* CTime#initialize */
static VALUE
ct_init (int argc, VALUE * argv, VALUE self)
{
  VALUE server, user, passwd;

  switch (rb_scan_args (argc, argv, "03", &server, &user, &passwd))
    {
    case 0:
      break;
    case 1:
    case 2:
      rb_raise (eCTime_CAPIError, "Must initialise with 0 or 3 arguments.");
    case 3:
      ct_connect (self, server, user, passwd);
    };

  return self;
}


/* CTime::Agenda#close */
static VALUE
ct_close_agenda (VALUE self)
{
  CAPISession *c_session = NULL;
  CAPIHandle *c_handle = NULL;
  CAPIStatus c_status = CAPI_STAT_OK;
  int i, j;

  if (rb_iv_get (self, "@user") == Qnil)
    {
      rb_ctime_debug ("Ignored attempt to close a closed agenda.\n");
      return Qnil;
    }

  rb_ctime_debug ("Attempting to close agenda for %s...\n",
		  STR2CSTR (rb_iv_get (self, "@user")));

  Data_Get_Struct (rb_iv_get (self, "@session"), CAPISession, c_session);
  Data_Get_Struct (self, CAPIHandle, c_handle);

  for (i = 1; i <= agenda_count; ++i)
    if (c_handle == &agenda[i - 1])
      {
	rb_ctime_debug ("Found matching agenda in global pool.\n");
	rb_ctime_debug ("Freeing the agenda handle...\n");

	/* close up hole in array */
	for (j = i - 1; j < agenda_count; ++j)
	  agenda[j] = agenda[j + 1];

	--agenda_count;

	c_status = capi_destroy_multiple_handles (*c_session, c_handle, 1);

	rb_ctime_debug ("Destroyed agenda.\n");
	ct_status_update (c_status);

	break;
      }

  ct_report_agendas ();

  rb_iv_set (self, "@user", Qnil);
  rb_iv_set (self, "@name", Qnil);
  rb_iv_set (self, "@session", Qnil);
  rb_iv_set (self, "@mailto", Qnil);
  rb_iv_set (self, "@type", Qnil);

  return Qnil;
}


/* CTime#open_agenda */
static VALUE
ct_open_agenda (VALUE self, VALUE user)
{
  const char *info = NULL;
  CAPISession *c_session = NULL;
  CAPIStatus c_status = CAPI_STAT_OK;
  VALUE new_agenda;

  if (agenda_count >= MAX_HANDLES)
    {
      rb_raise (eCTime_AgendaError, "Too many open agendas");
      return Qnil;
    }

  rb_ctime_debug ("Opening agenda for %s...\n", STR2CSTR (user));
  agenda[agenda_count] = NULL;

  rb_ctime_debug ("Getting handle...\n");
  Data_Get_Struct (self, CAPISession, c_session);
  c_status =
    capi_get_handle (*c_session, STR2CSTR (user), &agenda[agenda_count]);
  ct_status_update (c_status);

  /* No user on the server matched this name */
  if (c_status & CAPI_STAT_DATA_USERID)
    {
      agenda[agenda_count] = NULL;
      rb_raise (eCTime_AgendaError,
		"Agenda for %s does not exist", STR2CSTR (user));
    }

  if (c_status != CAPI_STAT_OK)
    {
      agenda[agenda_count] = NULL;
      rb_raise (eCTime_AgendaError, "Failed to get agenda handle");
    }

  rb_ctime_debug ("Creating agenda object...\n");
  new_agenda = Data_Wrap_Struct (cAgenda, NULL, NULL, &agenda[agenda_count]);
  rb_iv_set (new_agenda, "@session", self);
  rb_iv_set (new_agenda, "@user", user);

  rb_ctime_debug ("Getting agenda handle info...\n");
  c_status = capi_get_handle_info (*c_session, agenda[agenda_count],
				   CAPI_HANDLE_TYPE, &info);
  if (c_status != CAPI_STAT_OK)
    {
      if (agenda[agenda_count])
	capi_destroy_multiple_handles (*c_session, &agenda[agenda_count], 1);

      agenda[agenda_count] = NULL;
      rb_raise (eCTime_AgendaError, "Failed to get agenda type info");
    }

  rb_iv_set (new_agenda, "@type", rb_str_new2 (info));

  c_status = capi_get_handle_info (*c_session, agenda[agenda_count],
				   CAPI_HANDLE_NAME, &info);
  if (c_status != CAPI_STAT_OK)
    {
      if (agenda[agenda_count])
	capi_destroy_multiple_handles (*c_session, &agenda[agenda_count], 1);

      agenda[agenda_count] = NULL;
      rb_raise (eCTime_AgendaError, "Failed to get agenda name info");
    }

  rb_iv_set (new_agenda, "@name", rb_str_new2 (info));

  c_status = capi_get_handle_info (*c_session, agenda[agenda_count],
				   CAPI_HANDLE_MAILTO, &info);
  if (c_status != CAPI_STAT_OK)
    rb_iv_set (new_agenda, "@mailto", Qnil);
  else
    rb_iv_set (new_agenda, "@mailto", rb_str_new2 (info));

  rb_ctime_debug ("Got all agenda handle data.\n");
  ++agenda_count;

  /* Null terminate the array for ct_get_events */
  agenda[agenda_count] = NULL;
  ct_report_agendas ();

  /* Yield the agenda to the block, if given. */
  if (rb_block_given_p ())
    rb_ensure (rb_yield, new_agenda, ct_close_agenda, new_agenda);

  return new_agenda;
}


/* CTime::Agenda#get_event */
static VALUE
ct_get_event (VALUE self, VALUE uid)
{
  const char *stream_data = NULL;
  const char *error = NULL;
  CAPISession *c_session = NULL;
  CAPIStream c_stream = NULL;
  CAPIStatus c_status = CAPI_STAT_OK;
  VALUE event_list;

  Check_Type (uid, T_STRING);

  rb_ctime_debug ("Preparing to get events for %s ...\n",
		  (RSTRING (uid)->ptr));

  /* Create a memory stream to receive the information */
  Data_Get_Struct (rb_iv_get (self, "@session"), CAPISession, c_session);
  ct_make_mem_stream (*c_session, &c_stream, NULL, &stream_data);

  rb_ctime_debug ("Fetching events...\n");
  c_status =
    capi_fetch_event_by_uid (*c_session, agenda[agenda_count - 1],
			     RSTRING (uid)->ptr, NULL, 0, c_stream);

  error = ct_status_update (c_status);

  if ((c_status != CAPI_STAT_OK) && (c_status != CAPI_STAT_SECUR_READ_PROPS))
    {
      /* Release the stream */
      ct_kill_mem_stream (*c_session, &c_stream);
      rb_raise (eCTime_CAPIError, "Failed to fetch events: %s", error);
    }

  rb_ctime_debug ("Calling parse_events...\n");
  event_list = rb_funcall (cEvent, rb_intern ("parse_events"),
			   1, rb_str_new2 (stream_data));

  /* Release the stream */
  ct_kill_mem_stream (*c_session, &c_stream);

  /* Yield each event in list to block, if there is one. */
  if (rb_block_given_p () /* && TYPE(event_list) == T_ARRAY */ )
    rb_iterate (rb_each, event_list, rb_yield, Qnil);

  rb_ctime_debug ("get_event() finished.\n");

  return event_list;
}


/* CTime::Agenda#get_events */
static VALUE
ct_get_events (int argc, VALUE * argv, VALUE self)
{
  const char *stream_data = NULL;
  const char *error = NULL;
  CAPISession *c_session = NULL;
  CAPIStream c_stream = NULL;
  CAPIStatus c_status = CAPI_STAT_OK;
  VALUE start_time;
  VALUE end_time;
  VALUE should_parse;
  VALUE event_list;

  rb_scan_args (argc, argv, "21", &start_time, &end_time, &should_parse);
  if (should_parse == Qnil)
    should_parse = Qtrue;

  rb_ctime_debug ("Preparing to get events...\n");

  /* Check whether start_time or end_time responds to getutc. If so,
     convert values to UTC. */
  if (rb_respond_to (start_time, rb_intern ("getutc")) != 0)
    {
      start_time = rb_funcall (start_time, rb_intern ("getutc"), 0);
    }
  else if (rb_respond_to (start_time, rb_intern ("new_offset")) != 0)
    {
      start_time = rb_funcall (start_time,
			       rb_intern ("new_offset"), 1, INT2NUM (0));
    }

  if (rb_respond_to (end_time, rb_intern ("getutc")) != 0)
    {
      end_time = rb_funcall (end_time, rb_intern ("getutc"), 0);
    }
  else if (rb_respond_to (end_time, rb_intern ("new_offset")) != 0)
    {
      end_time = rb_funcall (end_time,
			     rb_intern ("new_offset"), 1, INT2NUM (0));
    }

  /* Check whether start_time or end_time responds to strftime. If so,
     convert it to a String. */
  if (rb_respond_to (start_time, rb_intern ("strftime")) != 0)
    {
      start_time = rb_funcall (start_time, rb_intern ("strftime"),
			       1, rb_str_new2 (CTIME_STRFTIME));
    }
  if (rb_respond_to (end_time, rb_intern ("strftime")) != 0)
    {
      end_time = rb_funcall (end_time, rb_intern ("strftime"),
			     1, rb_str_new2 (CTIME_STRFTIME));
    }

  rb_ctime_debug ("Will fetch events in range %s to %s...\n",
		  STR2CSTR (start_time), STR2CSTR (end_time));

  /* create a memory stream to receive the information */
  Data_Get_Struct (rb_iv_get (self, "@session"), CAPISession, c_session);
  ct_make_mem_stream (*c_session, &c_stream, NULL, &stream_data);

  rb_ctime_debug ("Fetching events...\n");
  c_status = capi_fetch_events_by_range (*c_session, agenda, agenda_count,
					 STR2CSTR (start_time),
					 STR2CSTR (end_time), NULL, 0,
					 c_stream);
  error = ct_status_update (c_status);

  /* 
   * The API returns CAPI_STAT_SECUR_READ_PROPS to indicate that some of the
   * requested properties weren't returned (i.e. there are private meetings).
   * It's OK, though - the stream is still populated with only the readable
   * attributes for the private meetings and all the attributes for public
   * ones.
   */
  if ((c_status != CAPI_STAT_OK) && (c_status != CAPI_STAT_SECUR_READ_PROPS))
    {
      /* Release the stream */
      ct_kill_mem_stream (*c_session, &c_stream);
      rb_raise (eCTime_CAPIError, "Failed to fetch events: %s", error);
    }

  if (should_parse == Qtrue)
    {
      rb_ctime_debug ("Calling parse_events...\n");
      event_list = rb_funcall (cEvent, rb_intern ("parse_events"),
			       1, rb_str_new2 (stream_data));
    }
  else
    {
      rb_ctime_debug ("Returning raw iCal stream...\n");
      event_list = rb_str_new2 (stream_data);
    }

  /* Release the stream */
  ct_kill_mem_stream (*c_session, &c_stream);

  /* Yield each event in list to block, if there is one. */
  if (rb_block_given_p () /* && TYPE(event_list) == T_ARRAY */ )
    rb_iterate (rb_each, event_list, rb_yield, Qnil);

  rb_ctime_debug ("get_events() finished.\n");

  return event_list;
}


/* CTime#designate */
static VALUE
ct_designate (VALUE self, VALUE user)
{
  CAPISession *c_session = NULL;
  CAPIStatus c_status = CAPI_STAT_OK;
  const char *error = NULL;

  rb_ctime_debug ("Setting identity to %s...\n", STR2CSTR (user));
  Data_Get_Struct (self, CAPISession, c_session);
  c_status = capi_set_identity (*c_session, STR2CSTR (user));

  if (c_status & CAPI_STAT_SECUR_LOGON_AUTH)
    {
      /* This user has not granted you designate rights! */
      rb_raise (eCTime_DesignationError,
		"Insufficient designation rights for user %s",
		STR2CSTR (user));
    }
  else if (c_status & CAPI_STAT_DATA_USERID)
    {
      /* No user on the server matched this name! */
      rb_raise (eCTime_DesignationError, "User %s unknown", STR2CSTR (user));
    }

  error = ct_status_update (c_status);

  if (c_status != CAPI_STAT_OK)
    rb_raise (eCTime_DesignationError, "Failed to designate for user %s: %s",
	      STR2CSTR (user), error);

  return Qtrue;
}


static VALUE
ct_update_event (VALUE self, VALUE event,
		 const char *uid, const char *recur_id, int flags)
{
  CAPIStatus c_status = CAPI_STAT_OK;
  CAPISession *c_session = NULL;
  CAPIStream c_stream = NULL;
#if defined(HAVE_CAPI_90)
  const char *email = ct_get_strprop (self, "@mailto");
#endif
  const char *strprop = NULL;
  const char *error = NULL;
  char ical_event[32768];
  char new_event[32768];
  char tmp_buffer[32768];
  char uid_buffer[128];

  rb_ctime_debug ("Creating a new event on the server...\n");

  if (rb_iv_get (event, "@private") == Qfalse)
    sprintf (new_event, "CLASS:PUBLIC\r\n");
  else
    sprintf (new_event, "CLASS:PRIVATE\r\n");

  /* Assume start time, duration are ok since ctor requires valid ones */
  sprintf (tmp_buffer, "DTSTART:%s\r\n",
	   STR2CSTR (rb_funcall (event, rb_intern ("start_time_str"), 0)));
  strcat (new_event, tmp_buffer);

  sprintf (tmp_buffer, "DURATION:%s\r\n",
	   STR2CSTR (rb_funcall (event, rb_intern ("duration_str"), 0)));
  strcat (new_event, tmp_buffer);

  if (uid != NULL)
    {
      strncpy (uid_buffer, uid, sizeof (uid_buffer));

      if (0 == (flags & RBCTIME_DELETE))
	{
	  sprintf (tmp_buffer, "UID:%s\r\n", uid);
	  strcat (new_event, tmp_buffer);
	}
    }

#if defined(HAVE_CAPI_20) || defined(HAVE_CAPI_25)
  /* Setting this with the CSDK causes duplicates */
  if (recur_id != NULL)
    {
      sprintf (tmp_buffer, "RECURRENCE-ID:%s\r\n", recur_id);
      strcat (new_event, tmp_buffer);
    }
#endif

  if ((strprop = ct_get_strprop (event, "@summary")) != NULL)
    {
      sprintf (tmp_buffer, "SUMMARY:%s\r\n",
	       STR2CSTR (rb_funcall (event, rb_intern ("summary_str"), 0)));
      strcat (new_event, tmp_buffer);
    }

  if ((strprop = ct_get_strprop (event, "@description")) != NULL)
    {
      sprintf (tmp_buffer, "DESCRIPTION:%s\r\n",
	       STR2CSTR (rb_funcall
			 (event, rb_intern ("description_str"), 0)));
      strcat (new_event, tmp_buffer);
    }

  if ((strprop = ct_get_strprop (event, "@location")) != NULL)
    {
      sprintf (tmp_buffer, "LOCATION:%s\r\n",
	       STR2CSTR (rb_funcall (event, rb_intern ("location_str"), 0)));
      strcat (new_event, tmp_buffer);
    }

  if ((strprop = ct_get_strprop (event, "@organizer")) != NULL)
    {
      sprintf (tmp_buffer, "ORGANIZER:mailto:%s\r\n", strprop);
      strcat (new_event, tmp_buffer);
    }

#if defined(HAVE_CAPI_90)
  if (rb_iv_get (event, "@attendees") != Qnil)
    {
      VALUE attendees = rb_iv_get (event, "@attendees");
      VALUE a = Qnil;

      long index = 0;
      char *buf_ptr = tmp_buffer;
      const char *mailto = NULL;
      const char *cutype = NULL;

      while ((a = rb_ary_entry (attendees, index++)) != Qnil)
	{
	  mailto = ct_get_strprop (a, "@mailto");
	  cutype = ct_get_strprop (a, "@cutype");

	  /* If we're not importing, only add current user */
	  if ((0 == (flags & RBCTIME_IMPORT)) &&
	      (NULL == email || NULL == mailto || strcmp (mailto, email)))
	    continue;

	  buf_ptr += sprintf (buf_ptr, "ATTENDEE");

	  if (mailto == NULL)
	    mailto = "unknown@nodomain";

	  if (cutype != NULL)
	    buf_ptr += sprintf (buf_ptr, ";CUTYPE=3D%s", cutype);

	  if ((strprop = ct_get_strprop (a, "@partstat")) != NULL)
	    buf_ptr += sprintf (buf_ptr, ";PARTSTAT=3D%s", strprop);

	  if ((strprop = ct_get_strprop (a, "@role")) != NULL)
	    buf_ptr += sprintf (buf_ptr, ";ROLE=3D%s", strprop);

	  if ((strprop = ct_get_strprop (a, "@guid")) != NULL)
	    buf_ptr += sprintf (buf_ptr, ";X-ORACLE-GUID=3D%s", strprop);

	  if ((strprop = ct_get_strprop (a, "@cn")) != NULL)
	    buf_ptr += sprintf (buf_ptr, ";CN=3D%s", strprop);

	  buf_ptr += sprintf (buf_ptr, ":mailto:%s\r\n", mailto);

	  strcat (new_event, tmp_buffer);
	  buf_ptr = tmp_buffer;
	}
    }
#endif

  snprintf (tmp_buffer, sizeof (tmp_buffer), ICAL_TEMPLATE, new_event);
  snprintf (ical_event, sizeof (ical_event), MIME_HEADER, tmp_buffer);

  rb_ctime_debug ("Message looks like:\n%s\n", ical_event);

  /* Store the event */
  Data_Get_Struct (rb_iv_get (self, "@session"), CAPISession, c_session);
  ct_make_mem_stream (*c_session, &c_stream, ical_event, NULL);

  rb_ctime_debug ("Storing event...\n");
  c_status = capi_store_event (*c_session, agenda, agenda_count, c_stream,
			       uid_buffer, sizeof (uid_buffer),
			       recur_id, flags);

  error = ct_status_update (c_status);

  if (c_status != CAPI_STAT_OK)
    {
      ct_kill_mem_stream (*c_session, &c_stream);
      rb_raise (eCTime_AgendaError, "Failed to set event: %s", error);
    }

  /* Event has been updated, state is synchronised */
  rb_iv_set (event, "@prev_start_time", Qnil);

  ct_kill_mem_stream (*c_session, &c_stream);
  rb_ctime_debug ("Got UID %s ...\n", uid_buffer);

  return rb_str_new2 (uid_buffer);
}


/* CTime::Agenda#set_event */
static VALUE
ct_set_event (VALUE self, VALUE event)
{
  const char *uid = ct_get_strprop (event, "@uid");
  const char *prev_start = ct_get_strprop (event, "@prev_start_time");

#if defined(HAVE_CAPI_90)
  VALUE uid_val = Qnil;

  if (NULL != uid)
    {
      int flag1 = 0;
      int flag2 = 0;

      if (NULL != prev_start)
	{
	  /* start time has changed; replace old event with new one */
	  flag1 = RBCTIME_REPLACE;
	  flag2 = RBCTIME_REPLY;
	}
      else
	{
	  /* start time unchanged; can overwrite old event */
	  flag1 = RBCTIME_IMPORT;
	  flag2 = RBCTIME_MODIFY;
	}

      /* use uid from first update for second update */
      uid_val = ct_update_event (self, event, uid, prev_start, flag1);
      uid = uid_val == Qnil ? uid : STR2CSTR (uid_val);

      return ct_update_event (self, event, uid, prev_start, flag2);
    }
  else
    {
      uid_val = ct_update_event (self, event, NULL, NULL, RBCTIME_IMPORT);

      if (uid_val == Qnil)
	rb_raise (eCTime_AgendaError, "Initial event import failed");

      /* reply stage is required to get owner's attendance set correctly */
      uid = STR2CSTR (uid_val);
      return ct_update_event (self, event, uid, NULL, RBCTIME_REPLY);
    }
#endif

  return ct_update_event (self, event, uid, prev_start, RBCTIME_NULL);
}


#if defined(HAVE_CAPI_90)
/* CTime::Agenda#reply */
static VALUE
ct_reply (VALUE self, VALUE event)
{
  const char *uid = ct_get_strprop (event, "@uid");
  const char *prev_start = ct_get_strprop (event, "@prev_start_time");

  return ct_update_event (self, event, uid, prev_start, RBCTIME_REPLY);
}
#endif


/* CTime::Agenda#delete_event */
static VALUE
ct_delete_event (int argc, VALUE * argv, VALUE self)
{
  CAPIStatus c_status = CAPI_STAT_OK;
  CAPISession *c_session = NULL;
  VALUE uid_arg = Qnil;
  VALUE recur_arg = Qnil;
  const char *uid = NULL;
  const char *recur = NULL;

  rb_ctime_debug ("Deleting event from server...\n");

  rb_scan_args (argc, argv, "11", &uid_arg, &recur_arg);

  uid = STR2CSTR (uid_arg);

  if (recur_arg != Qnil)
    {
      /* Check whether recur_arg responds to getutc. If so,
         convert the value to UTC. */
      if (rb_respond_to (recur_arg, rb_intern ("getutc")) != 0)
	{
	  recur_arg = rb_funcall (recur_arg, rb_intern ("getutc"), 0);
	}
      else if (rb_respond_to (recur_arg, rb_intern ("new_offset")) != 0)
	{
	  recur_arg =
	    rb_funcall (recur_arg, rb_intern ("new_offset"), 1, INT2NUM (0));
	}

      /* Check whether recur_arg responds to strftime. If so,
         convert it to a String. */
      if (rb_respond_to (recur_arg, rb_intern ("strftime")) != 0)
	{
	  recur_arg = rb_funcall (recur_arg, rb_intern ("strftime"),
				  1, rb_str_new2 (CTIME_STRFTIME));
	}

      recur = STR2CSTR (recur_arg);
    }

  rb_ctime_debug ("UID is %s\n", uid);
  rb_ctime_debug ("RECURRENCE-ID is %s\n", recur ? recur : "(nil)");

  Data_Get_Struct (rb_iv_get (self, "@session"), CAPISession, c_session);
  c_status = capi_delete_event (*c_session, agenda, agenda_count, uid, recur);
  ct_status_update (c_status);

  if (c_status == CAPI_STAT_DATA_UID_NOTFOUND)
    rb_raise (eCTime_AgendaError, "No matching event for UID %s", uid);

  if (c_status == CAPI_STAT_DATA_UID_RECURRENCE)
    rb_raise (eCTime_AgendaError, "No matching event for UID %s at %s",
	      uid, recur);

  if (c_status != CAPI_STAT_OK)
    rb_raise (eCTime_AgendaError, "Failed to delete event");

  return Qtrue;
}


/* Initialize the class */
void
Init_ctime ()
{
  rb_ctime_debug ("Initialising Ruby/CorporateTime library...\n");

  /* Define the CTime class */
  cCTime = rb_define_class ("CTime", rb_cObject);

  /* Define the CTime::Agenda class */
  cAgenda = rb_define_class_under (cCTime, "Agenda", rb_cObject);
  rb_ctime_debug ("Requiring agenda.rb...\n");
  rb_require ("ctime/agenda");

  /* Define the CTime::Agenda::Event class */
  cEvent = rb_define_class_under (cAgenda, "Event", rb_cObject);
  rb_ctime_debug ("Requiring event.rb...\n");
  rb_require ("ctime/event");

  /* Require attendee.rb */
  rb_ctime_debug ("Requiring attendee.rb...\n");
  rb_require ("ctime/attendee");

  /* Define the CTime::CTimeError exception */
  eCTime_CAPIError = rb_define_class_under (cCTime, "CAPIError",
					    rb_eStandardError);

  /* Define the CTime::AgendaError exception */
  eCTime_AgendaError = rb_define_class_under (cAgenda, "AgendaError",
					      rb_eStandardError);

  /* Define the CTime::DesignationError exception */
  eCTime_DesignationError =
    rb_define_class_under (cCTime, "DesignationError", rb_eStandardError);

  /* Define CTime::VERSION constant */
  rb_define_const (cCTime, "VERSION", rb_str_new2 (CTIME_VERSION));

  /* Define CTime::STRFTIME constant */
  rb_define_const (cCTime, "STRFTIME", rb_str_new2 (CTIME_STRFTIME));

  /* Define the CTime#initialize method */
  rb_define_alloc_func (cCTime, ct_allocate);
  rb_define_method (cCTime, "initialize", ct_init, -1);
  rb_define_method (cCTime, "connect", ct_connect, 3);
  rb_define_method (cCTime, "capabilities", ct_capabilities, 0);

  /* Define the CTime#quit method */
  rb_define_method (cCTime, "quit", ct_quit, 0);

  /* Define the CTime#open_agenda method */
  rb_define_method (cCTime, "open_agenda", ct_open_agenda, 1);

  /* Define the CTime#designate method */
  rb_define_method (cCTime, "designate", ct_designate, 1);

  /* Define the CTime::Agenda#get_event method */
  rb_define_method (cAgenda, "get_event", ct_get_event, 1);

  /* Define the CTime::Agenda#get_events method */
  rb_define_method (cAgenda, "get_events", ct_get_events, -1);

  /* Define the CTime::Agenda#set_event method */
  rb_define_method (cAgenda, "set_event", ct_set_event, 1);

#if defined(HAVE_CAPI_90)
  /* Define the CTime::Agenda#reply method */
  rb_define_method (cAgenda, "reply", ct_reply, 1);
#endif

  /* Define the CTime::Agenda#delete_event method */
  rb_define_method (cAgenda, "delete_event", ct_delete_event, -1);

  /* Define the CTime::Agenda#close method */
  rb_define_method (cAgenda, "close", ct_close_agenda, 0);

  return;
}
